# Copyright 2021 MosaicML. All Rights Reserved.

import logging
import re
from abc import ABC
from dataclasses import asdict, dataclass, fields
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import torch
import yahp as hp
from torch.optim.lr_scheduler import CosineAnnealingLR, ExponentialLR, MultiStepLR, StepLR, _LRScheduler

from composer.core.types import Optimizer, Scheduler
from composer.optim.pytorch_future import WarmUpLR

log = logging.getLogger(__name__)

Time = str
"""
Time: For scheduler hparams, we support providing time (e.g. milestones) as
both integers, which will be interpreted as epochs, or as a string in format:
* '12ep' -- 12 epochs
* '1024ba' -- 1024 batches
* '12ep32ba' -- 12 epochs and 32 batches

The provided time is converted and represented internally.
"""

_interval_doc = 'frequency of step() calls, either "batch" or "epoch". Default: "epoch"'

STR_REGEX = re.compile(r'^(?:([0-9]*)(ep))?(?:([0-9]*)(ba))?$', flags=re.IGNORECASE)

# Allow (batch, batches) or (epoch, epochs). Also accept "step" ~ "batch"
INTERVAL_MAP = {
    'batch': 'batch',
    'batches': 'batch',
    'epoch': 'epoch',
    'epochs': 'epoch',
    'step': 'batch',
    'steps': 'batch'
}


def parse_time_string(timestring: str) -> Tuple[int, int]:
    """ Parse timestring to (epoch, batches).

    Args:
        timestring (str): String in the format 'XXepYYba'.

    Returns:
        tuple: (epochs, batches)

    Raises:
        ValueError: The timestring is invalid

    Examples:
        >>> parse_time_string('32ep173ba')
        (32, 173)
        >>> parse_time_string('12ep')
        (12, 0)
        >>> parse_time_string('1024ba')
        (0, 1024)
    """

    match = STR_REGEX.findall(timestring)
    if len(match) != 1:
        raise ValueError(f'Invalid timestring: {timestring}. Should be of format 32ep15ba, or 99ba or 7ep')
    match = match[0]

    epochs = 0 if 'ep' not in match else int(match[match.index('ep') - 1])
    batches = 0 if 'ba' not in match else int(match[match.index('ba') - 1])

    return epochs, batches


def _convert_time(time: Time, steps_per_epoch: Optional[int] = None, interval: str = 'epoch') -> int:
    """ Convert time to either batches or epochs (based on interval argument) """
    if isinstance(time, int):
        return time
    if steps_per_epoch is None:
        raise ValueError('steps_per_epoch must be provided to parse time string.')

    epochs, batches = parse_time_string(time)
    if interval in ('batches', 'batch', 'steps', 'step'):
        log.info(f'Converting {time}, {interval} to {batches + epochs * steps_per_epoch}')
        return batches + epochs * steps_per_epoch
    elif interval in ('epochs', 'epoch'):
        epochs = epochs + batches // steps_per_epoch
        batches = batches % steps_per_epoch
        if batches != 0:
            log.warning('Scheduler is stepping every epoch, but provided timestring '
                        f'{time} had batches. Ignoring the batches term.')
        log.info(f'Converting {time}, {interval} to {epochs}')
        return epochs
    else:
        raise ValueError('interval must be one of (batch, epoch)')


@dataclass
class SchedulerHparams(hp.Hparams, ABC):

    scheduler_object = None  # type: Optional[Callable[..., Scheduler]]
    interval = 'epochs'  # type: str

    def convert_time_fields(self, steps_per_epoch: Optional[int] = None) -> None:
        """Converts all fields that were provided as timestrings (e.g. "32ep11ba") into
        integers, representing either epochs or batches, depending on the
        scheduler's interval attribute.

        Examples:
            >>> hp = StepLRHparams(step_size='32ep77ba', interval='batch')
            >>> hp.convert_time_fields(steps_per_epoch=100)
            >>> hp.step_size
            3277
            >>> hp = StepLRHparams(step_size='32ep77ba', interval='epoch')
            >>> hp.convert_time_fields(steps_per_epoch=100)
            >>> hp.step_size
            32
            >>> hp = StepLRHparams(step_size=5, interval='epoch')
            >>> hp.convert_time_fields()  # steps_per_epoch not needed
            >>> hp.step_size
            5
            >>> hp = MultiStepLRHParams(milestones=['50ep', '8050ba'], interval='batch')
            >>> hp.convert_time_fields(steps_per_epoch=100)
            >>> hp.milestones
            [5000, 8050]
            >>> hp = MultiStepLRHParams(milestones=['50ep', '8050ba'], interval='epoch')
            >>> hp.convert_time_fields(steps_per_epoch=100)
            >>> hp.milestones
            [50, 80]

        Args:
            steps_per_epoch (int): used to convert between epochs <-> batches. Need not be
                                   provided if all fields are provided as integers.
        """
        assert hasattr(self, 'interval'), "Scheduler Hparams needs an interval (str) parameter."

        for field in fields(self):
            # TODO: switch Time back to Union[int, str]
            if field.name not in ('interval', 'warmup_method') and field.type == Time or field.type == List[Time]:
                time = getattr(self, field.name)
                if isinstance(time, list):
                    result = [_convert_time(t, steps_per_epoch, self.interval) for t in time]
                else:
                    result = _convert_time(time, steps_per_epoch, self.interval)

                setattr(self, field.name, result)

    def initialize_object(  # type: ignore
            self,
            optimizer: Optimizer,
            steps_per_epoch: Optional[int] = None,
    ) -> Tuple[Scheduler, str]:
        """ Create the scheduler object from the current hparams

        Args:
            optimizer (Optimizer): the optimizer associated with this scheduler
            steps_per_epoch (Optional[int], optional): number of steps per epoch. Default: ``None``.

        Returns:
            (Scheduler, str): (The parametrized scheduler instance, schedule step interval)
        """

        assert self.scheduler_object is not None, "Scheduler Hparams needs scheduler_object to initialize."
        assert hasattr(self, 'interval'), "Scheduler Hparams needs an interval (str) parameter."
        self.convert_time_fields(steps_per_epoch)

        # we pass the interval to the trainer directly
        kwargs = {k: v for k, v in asdict(self).items() if k not in ['interval']}
        obj = self.scheduler_object(optimizer, **kwargs)
        obj.interval = self.interval  # type: ignore
        obj.steps_per_epoch = steps_per_epoch  # type: ignore
        return obj, self.interval


class ConstantLR(_LRScheduler):
    """ Scheduler that does not change the optimizer's learning rate.

    Args:
        optimizer (Optimizer): the optimizer associated with this scheduler.
        last_epoch (int, optional): The index of the last epoch. Can be used to restore the state of the
                                    learning rate schedule. Default: ``-1``.
        verbose (bool, optional): If ``True``, prints a message to stdout for each update. Default: ``False``.
    """

    def __init__(self, optimizer: Optimizer, last_epoch: int = -1, verbose: int = False):

        self.optimizer = optimizer
        super(ConstantLR, self).__init__(optimizer, last_epoch, verbose)  # type: ignore

    def get_lr(self):
        """ Get the current learning rate for each parameter group.

        Returns:
            List of float: The current learning rate for each parameter group.
        """
        return self.base_lrs  # type: ignore

    def _get_closed_form_lr(self):
        """ Get the current learning rate for each parameter group.

        Returns:
            List of float: The current learning rate for each parameter group.
        """
        return [base_lr for base_lr in self.base_lrs]  # type: ignore


@dataclass
class ConstantLRHparams(SchedulerHparams):
    """ Hyperparameters for the :class:`ConstantLR` scheduler.
    """
    verbose: bool = hp.optional(default=False, doc='prints message to stdout')
    interval: str = hp.optional(default='epoch', doc=_interval_doc)

    scheduler_object = ConstantLR


@dataclass
class StepLRHparams(SchedulerHparams):
    """ Hyperparameters for the `StepLR <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.StepLR.html#torch.optim.lr_scheduler.StepLR>`_
    scheduler.
    """

    step_size: Time = hp.required(doc='Period of learning rate decay')
    gamma: float = hp.optional(default=0.1, doc='multiplicative factor of decay')
    verbose: bool = hp.optional(default=False, doc='prints message to stdout')
    interval: str = hp.optional(default='epoch', doc=_interval_doc)

    scheduler_object = torch.optim.lr_scheduler.StepLR


@dataclass
class MultiStepLRHparams(SchedulerHparams):
    """ Hyperparameters for the `MultiStepLR <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.MultiStepLR.html#torch.optim.lr_scheduler.MultiStepLR>`_
    scheduler.
    """

    milestones: List[Time] = hp.required(doc='List of epoch indicies')
    gamma: float = hp.optional(default=0.1, doc='multiplicative factor of decay')
    verbose: bool = hp.optional(default=False, doc='prints message to stdout')
    interval: str = hp.optional(default='epoch', doc=_interval_doc)

    scheduler_object = torch.optim.lr_scheduler.MultiStepLR


@dataclass
class ExponentialLRHparams(SchedulerHparams):
    """ Hyperparameters for the `ExponentialLR <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.ExponentialLR.html#torch.optim.lr_scheduler.ExponentialLR>`_
    scheduler.
    """

    gamma: float = hp.required(doc='multiplicative factor of decay')
    verbose: bool = hp.optional(default=False, doc='prints message to stdout')
    interval: str = hp.optional(default='epoch', doc=_interval_doc)

    scheduler_object = torch.optim.lr_scheduler.ExponentialLR


@dataclass
class CosineAnnealingLRHparams(SchedulerHparams):
    """ Hyperparameters for the `CosineAnnealingLR <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingLR.html#torch.optim.lr_scheduler.CosineAnnealingLR>`_
    scheduler.
    """

    T_max: Time = hp.required(doc="Maximum number of iterations.")
    eta_min: float = hp.optional(default=0.0, doc='minimum learning rate.')
    verbose: bool = hp.optional(default=False, doc='prints message to stdout')
    interval: str = hp.optional(default='epoch', doc=_interval_doc)

    scheduler_object = torch.optim.lr_scheduler.CosineAnnealingLR

    def initialize_object(self, optimizer: Optimizer, steps_per_epoch: Optional[int] = None):
        self.convert_time_fields(steps_per_epoch)
        return super().initialize_object(optimizer, steps_per_epoch)


@dataclass
class CosineAnnealingWarmRestartsHparams(SchedulerHparams):
    """ Hyperparameters for the ``CosineAnnealingWarmRestarts` <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingWarmRestarts.html#torch.optim.lr_scheduler.CosineAnnealingWarmRestarts>`_
    scheduler.
    """

    T_0: Time = hp.required("Number of iterations for the first restart.")
    eta_min: float = hp.optional(default=0.0, doc='minimum learning rate.')
    verbose: bool = hp.optional(default=False, doc='prints message to stdout')
    interval: str = hp.optional(default='epoch', doc=_interval_doc)
    T_mult: int = hp.optional("A factor increases :math:`T_{i}` after a restart. Default: 1.", default=1)

    scheduler_object = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts

    def initialize_object(self, optimizer: Optimizer, steps_per_epoch: Optional[int] = None):
        self.convert_time_fields(steps_per_epoch)
        return super().initialize_object(optimizer, steps_per_epoch)


@dataclass
class WarmUpLRHparams(SchedulerHparams):
    """ Hyperparameters for the :class:`~composer.optim.pytorch_future.WarmUpLR` scheduler.

    See the documentation for :class:`~composer.optim.pytorch_future.WarmUpLR`.
    """

    warmup_factor: float = hp.optional("Number to multiply learning rate at start.", default=1.0 / 3)
    warmup_iters: Time = hp.optional("Number of warmup step. Default: 5 iterations.", default="5ba")
    warmup_method: str = hp.optional("Warmup method (linear or constant)", default='linear')
    verbose: bool = hp.optional('Prints message to stdout', default=False)
    interval: str = hp.optional('Warmup the LR every step or epoch. Default: epoch', default='epoch')

    scheduler_object = WarmUpLR


def ensure_warmup_last(schedulers: List[SchedulerHparams]) -> List[SchedulerHparams]:
    """ Ensure that WarmUp-based schedulers appear last in the provided list.

    Args:
        schedulers (List[SchedulerHparams]): List of schedulers.

    Returns:
        List[SchedulerHparams]: A sorted list of schedulers with WarmUp-based schedulers at the end.
    """

    return sorted(schedulers, key=lambda x: isinstance(x, (WarmUpLR, WarmUpLRHparams)))


def get_num_warmup_batches(scheduler_hparams: Sequence[SchedulerHparams], steps_per_epoch: Optional[int] = None) -> int:
    """ Gets the number of warmup steps declared by a list of schedulers.

    Args:
        scheduler_hparams (Sequence[SchedulerHparams]): List of schedulers
        steps_per_epoch (Optional[int], optional): Number of steps in a single epoch. Default: ``None``.

    Returns:
        int: Number of warmup steps
    """

    warmup_scheduler_hparams = [scheduler for scheduler in scheduler_hparams if isinstance(scheduler, WarmUpLRHparams)]
    if len(warmup_scheduler_hparams):
        warmup_iters = warmup_scheduler_hparams[0].warmup_iters
        if isinstance(warmup_iters, str):
            return _convert_time(
                time=warmup_iters,
                steps_per_epoch=steps_per_epoch,
                interval=warmup_scheduler_hparams[0].interval,
            )
        else:
            return warmup_iters
    return 0


class ComposedScheduler(_LRScheduler):
    """Handles warmup for a chained list of schedulers.

    With one call, will run each scheduler's ``step()``. If :class:`WarmUpLR` is in the list, will delay the stepping of
    schedulers that need to be silent during warmup. ``ComposedScheduler`` handles warmups, where as `ChainedScheduler <https://pytorch.org/docs/1.10./generated/torch.optim.lr_scheduler.ChainedScheduler.html?highlight=chained#torch.optim.lr_scheduler.ChainedScheduler>`_
    only combines schedulers.

    `CosineAnnealingLR <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingLR.html#torch.optim.lr_scheduler.CosineAnnealingLR>`_
    and `ExponentialLR <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.ExponentialLR.html#torch.optim.lr_scheduler.ExponentialLR>`_
    are not stepped during the warmup period. Other schedulers, such as
    `MultiStepLR <https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.MultiStepLR.html#torch.optim.lr_scheduler.MultiStepLR>`_
    are still stepped, to keep their milestones unchanged.

    Handles running the :class:`WarmUpLR` at every step if :attr:`WarmUpLR.interval='batch'`, and other schedulers at
    every epoch.

    Args:
        schedulers (list): List of chained schedulers.
    Example:
        >>> # Assuming optimizer uses lr = 1. for all groups
        >>> # lr = 0.1      if epoch == 0
        >>> # lr = 0.1      if epoch == 1
        >>> # lr = 0.9      if epoch == 2  # ExponentialLR effect starts here
        >>> # lr = 0.81     if epoch == 3
        >>> # lr = 0.729    if epoch == 4
        >>> scheduler1 = WarmUpLR(self.opt, warmup_factor=0.1, warmup_iters=2, warmup_method="constant")
        >>> scheduler2 = ExponentialLR(self.opt, gamma=0.9)
        >>> scheduler = ComposedScheduler(zip([scheduler1, scheduler2], ["epoch", "epoch"]))
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()

        >>> # Assuming optimizer uses lr = 1. for all groups
        >>> # lr = 0.1      if epoch == 0
        >>> # lr = 0.1      if epoch == 1
        >>> # lr = 1.0      if epoch == 2
        >>> # lr = 1.0     if epoch == 3
        >>> # lr = 0.2    if epoch == 4 . # MultiStepLR effect starts here
        >>> scheduler1 = WarmUpLR(self.opt, warmup_factor=0.1, warmup_iters=2, warmup_method="constant")
        >>> scheduler2 = MultiStepLR(optimizer, milestones=[4], gamma=0.2)
        >>> scheduler = ComposedScheduler(zip([scheduler1, scheduler2], ["epoch", "epoch"]))
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()
    """

    def __init__(self, schedulers):

        # check for tuple
        if not all(isinstance(scheduler, tuple) for scheduler in schedulers):
            raise ValueError('Schedulers must be a tuple of (Scheduler, interval), '
                             'where interval is one of "epoch" or "batch".')

        self._validate_same_optimizers(schedulers)
        self.schedulers, self.intervals = list(zip(*schedulers))  # unpack (scheduler, interval)

        # generous with spelling (batch, batches)/(step, steps) and (epoch, epochs)
        self.intervals = [INTERVAL_MAP[interval] for interval in self.intervals]

        warmup = [(scheduler, interval)
                  for scheduler, interval in zip(self.schedulers, self.intervals)
                  if isinstance(scheduler, WarmUpLR)]
        if warmup:
            assert len(warmup) == 1, "ComposedScheduler only supports one WarmUpLR " \
                                     f"in the provided list, found {len(warmup)}."
            warmup, interval = warmup[0]
            self.warmup_iters = warmup.warmup_iters
            log.info(f'Setting LR Warmup to {self.warmup_iters} {interval}')
        else:
            self.warmup_iters = 0

        # these schedulers need to be silent during warmup
        self.delay_schedulers = [CosineAnnealingLR, ExponentialLR]
        self._warmup_counter = 0  # counter to track warmups

    def step(self, interval: str = 'epoch'):
        """ Step all applicable schedulers.

        Args:
            interval (str, optional): The interval of the current step. Must be either ``'step'`` or ``'epoch'``.
                                      Default: ``epoch``.
        """
        for scheduler, scheduler_interval in zip(self.schedulers, self.intervals):
            if self._warmup_counter < self.warmup_iters and \
                any(isinstance(scheduler, delay) for delay in self.delay_schedulers):
                continue

            if interval == scheduler_interval:
                scheduler.step()
                if isinstance(scheduler, WarmUpLR):
                    self._warmup_counter += 1

    def _validate_schedulers(self, warmup_epochs: int) -> None:
        """ Verify that any stepwise schedulers do not change the LR during the desired warmup period.

        Args:
            warmup_epochs (int): Number of epochs for warmup.
        """
        # since WarmUpLR is non-chainable form, step LR milestones must
        # occur after warmup is completed
        lr_step_schedulers = [
            scheduler for scheduler in self.schedulers if isinstance(scheduler, (StepLR, MultiStepLR))
        ]
        for scheduler in lr_step_schedulers:
            if isinstance(scheduler, StepLR) and scheduler.step_size <= warmup_epochs:  # type: ignore
                raise ValueError(f'StepLR step_size {scheduler.step_size} must '  # type: ignore
                                 'be greater than warmup_iters {self.warmup_iters}')
            elif isinstance(scheduler, MultiStepLR):
                if any(ms <= warmup_epochs for ms in scheduler.milestones.elements()):  #type: ignore
                    raise ValueError(f'MultiStepLR milestones must be greater than warmup_iters {warmup_epochs}')

    def state_dict(self) -> Dict[str, Any]:
        """ Returns a dictionary containing the state of all composed schedulers.

        Returns:
            Dict: the state dictionary
        """
        state_dict = {
            "schedulers": {scheduler.__class__.__qualname__: scheduler.state_dict() for scheduler in self.schedulers},
            "_warmup_counter": self._warmup_counter,
        }
        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """ Load the state of all composed schedulers from the provided dictionary.

        Args:
            state_dict (Dict[str, Any]): A dict containing the state of all composed schedulers. Should be an object
            returned from a call to :meth:`state_dict()`.
        """
        for scheduler in self.schedulers:
            scheduler.load_state_dict(state_dict["schedulers"][scheduler.__class__.__qualname__])
        self._warmup_counter = state_dict["_warmup_counter"]

    def _validate_same_optimizers(self, schedulers):
        """ Verify that all schedulers correspond to the same optimizer
        """
        for scheduler_idx in range(1, len(schedulers)):
            if (schedulers[scheduler_idx][0].optimizer != schedulers[0][0].optimizer):  # type: ignore
                raise ValueError("ComposedScheduler expects all schedulers to belong to the same optimizer, but "
                                 "got schedulers at index {} and {} to be different".format(0, scheduler_idx))



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State

log = logging.getLogger(__name__)


@dataclass
class MyAlgorithmHparams(AlgorithmHparams):
    """
    This hparams object is for use with our hparams system for specifying algorithms via YAML
    and argument parser flags.
    """

    alpha: float = hp.optional(doc='optional hparams need a default field. This field would'
                               'not be required for CLI flags, but still required for the YAML config.',
                               default=0.1)
    beta: float = hp.required(doc='required fields need a template_default, which '
                              'is used to generate templates. '
                              'This field is still required for CLI flags.',
                              template_default=0.5)

    def initialize_object(self) -> MyAlgorithm:
        """
        **delete this comment before contributing**

        Factory method that links this hparams object with the algorithm class. For algorithms
        that require special imports, can only run the imports during calls to this function.

        e.g.
        ```
        from composer.algorithms.my_algorithm import MyAlgorithm
        ```
        """
        return MyAlgorithm(**asdict(self))


def my_algorithm(model, alpha: float, beta: float):
    """
    Most of your algorithm logic should be in a functional form here, similar to
    torch.nn.Conv2d vs torch.nn.functional.conv2d
    """
    return model


class MyAlgorithm(Algorithm):
    """
    Add docstrings in Google style. See:

    https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html
    """

    def __init__(self, alpha: float, beta: float = 0.5):
        """
        __init__ is constructed from the same fields as in hparams.
        """
        self.hparams = MyAlgorithmHparams(alpha, beta)

    def match(self, event: Event, state: State) -> bool:
        return True

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """
        Implement your algorithm's state change here.
        """
        state.model = my_algorithm(state.model, self.hparams.alpha, self.hparams.beta)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.resnet101.model import ResNet101 as ResNet101
from composer.models.resnet101.resnet101_hparams import ResNet101Hparams as ResNet101Hparams

_task = 'Image Classification'
_dataset = 'ImageNet'
_name = 'ResNet101'
_quality = ''
_metric = 'Top-1 Accuracy'
_ttt = '?'
_hparams = 'hparams.yaml'



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import inspect
from dataclasses import asdict, dataclass
from typing import Callable, Optional, Tuple

import numpy as np
import torch
import yahp as hp
from torch.nn import functional as F

from composer.algorithms.algorithm_hparams import AlgorithmHparams
from composer.core.types import Algorithm, Event, Logger, State, Tensor


def do_selective_backprop(
    epoch: int,
    batch_idx: int,
    start_epoch: int,
    end_epoch: int,
    interrupt: int,
) -> bool:
    """Decide if selective backprop should be run based on time in training.

    Returns true if the current `epoch` is between `start_epoch` and `end_epoch`.
    Recommend that SB be applied during the later stages of a training run,
    once the model has already "learned" easy examples.

    To preserve convergence, SB can be interrupted with vanilla minibatch
    gradient steps every `interrupt` steps. When `interrupt=0`, SB will be
    used at every step during the SB interval. When `interrupt=2`, SB will
    alternate with vanilla minibatch steps.

    Args:
        epoch (int): The current epoch during training.
        batch_idx (int): The current batch within the epoch.
        start_epoch (int): The epoch at which selective backprop should be enabled.
        end_epoch (int): The epoch at which selective backprop should be disabled.
        interrupt (int): The number of batches between vanilla minibatch gradient updates.

    Returns:
        bool: If selective backprop should be performed on this batch.
    """
    is_interval = ((epoch >= start_epoch) and (epoch < end_epoch))
    is_step = ((interrupt == 0) or ((batch_idx + 1) % interrupt != 0))

    return is_interval and is_step


def selective_backprop(X: torch.Tensor,
                       y: torch.Tensor,
                       model: torch.nn.Module,
                       loss_fun: Callable,
                       keep: float,
                       scale_factor: float = 1) -> Tuple[torch.Tensor, torch.Tensor]:
    """Select a subset of the batch on which to learn.

    Selective Backprop (SB) prunes minibatches according to the difficulty
    of the individual training examples, and only computes weight gradients
    over the pruned subset, reducing iteration time and speeding up training.
    The fraction of the minibatch that is kept for gradient computation is
    specified by the argument `0 <= keep <= 1`.

    See `Accelerating Deep Learning by Focusing on the Biggest Losers` <https://arxiv.org/abs/1910.00762>.

    To speed up SB's selection forward pass, the argument `scale_factor` can
    be used to downsample input image tensors. The full-sized inputs will still
    be used for the weight gradient computation.

    Args:
        X (torch.Tensor): Input tensor to prune.
        y (torch.Tensor): Output tensor to prune.
        model (torch.nn.Module): Model with which to predict outputs.
        loss_fun (Callable): Loss function of the form `loss(outputs, targets, reduction='none')`.
                             The function must take the keyword argument `reduction='none'`
                             to ensure that per-sample losses are returned.
        keep (float): Fraction of examples in the batch to keep.
        scale_factor (float, optional): Scale factor for downsampling input
                                        tensors. Downsampling requires the input
                                        tensor to be at least 3D. Default: 1 (no downsampling)

    Returns:
        (torch.Tensor, torch.Tensor): The pruned batch of inputs and targets

    Note:
    This function runs an extra forward pass through the model on the batch of data.
    If you are using a non-default precision, ensure that this forward pass
    runs in your desired precision. For example:
    ```
    with torch.cuda.amp.autocast(True):
        X_new, y_new = selective_backprop(X, y, model, loss_fun, keep, scale_factor)
    ```
    """
    INTERPOLATE_MODES = {3: "linear", 4: "bilinear", 5: "trilinear"}

    interp_mode = "bilinear"
    if scale_factor != 1:
        if X.dim() not in INTERPOLATE_MODES:
            raise ValueError(f"Input must be 3D, 4D, or 5D if `scale_factor` != 1, got {X.dim()}")
        interp_mode = INTERPOLATE_MODES[X.dim()]

    if scale_factor > 1:
        raise ValueError("`scale_factor` must be <= 1")

    if callable(loss_fun):
        sig = inspect.signature(loss_fun)
        if not "reduction" in sig.parameters:
            raise TypeError("Loss function `loss_fun` must take a keyword argument `reduction`.")
    else:
        raise TypeError("Loss function must be callable")

    with torch.no_grad():
        N = X.shape[0]

        # Maybe interpolate
        if scale_factor < 1:
            X_scaled = F.interpolate(X, scale_factor=scale_factor, mode=interp_mode)
        else:
            X_scaled = X

        # Get per-examples losses
        out = model(X_scaled)
        losses = loss_fun(out, y, reduction="none")

        # Sort losses
        sorted_idx = torch.argsort(losses)
        n_select = int(keep * N)

        # Sample by loss
        percs = np.arange(0.5, N, 1) / N
        probs = percs**((1.0 / keep) - 1.0)
        probs = probs / np.sum(probs)
        select_percs_idx = np.random.choice(N, n_select, replace=False, p=probs)
        select_idx = sorted_idx[select_percs_idx]

    return X[select_idx], y[select_idx]


@dataclass
class SelectiveBackpropHparams(AlgorithmHparams):

    start: float = hp.required(doc="SB interval start, as fraction of training duration", template_default=0.5)
    end: float = hp.required(doc="SB interval end, as fraction of training duration", template_default=0.9)
    keep: float = hp.required(doc="fraction of minibatch to select and keep for gradient computation",
                              template_default=0.5)
    scale_factor: float = hp.required(doc="scale for downsampling input for selection forward pass",
                                      template_default=0.5)
    interrupt: int = hp.required(doc="interrupt SB with a vanilla minibatch step every 'interrupt' batches",
                                 template_default=2)

    def initialize_object(self) -> SelectiveBackprop:
        return SelectiveBackprop(**asdict(self))


class SelectiveBackprop(Algorithm):
    """Selectively backprop on a subset of each batch.

    Selective Backprop (SB) prunes minibatches according to the difficulty
    of the individual training examples, and only computes weight gradients
    over the pruned subset, reducing iteration time and speeding up training.
    The fraction of the minibatch that is kept for gradient computation is
    specified by the argument `0 <= keep <= 1`.

    See `Accelerating Deep Learning by Focusing on the Biggest Losers` <https://arxiv.org/abs/1910.00762>.

    To speed up SB's selection forward pass, the argument `scale_factor` can
    be used to downsample input image tensors. The full-sized inputs will still
    be used for the weight gradient computation.

    To preserve convergence, SB can be interrupted with vanilla minibatch
    gradient steps every `interrupt` steps. When `interrupt=0`, SB will be
    used at every step during the SB interval. When `interrupt=2`, SB will
    alternate with vanilla minibatch steps.

    Args:
        start (float): SB interval start as fraction of training duration
        end (float): SB interval end as fraction of training duration
        keep (float): fraction of minibatch to select and keep for gradient computation
        scale_factor (float): scale for downsampling input for selection forward pass
        interrupt (int): interrupt SB with a vanilla minibatch step every 'interrupt' batches
    """

    def __init__(self, start: float, end: float, keep: float, scale_factor: float, interrupt: int):
        self.hparams = SelectiveBackpropHparams(start=start,
                                                end=end,
                                                keep=keep,
                                                scale_factor=scale_factor,
                                                interrupt=interrupt)

    def match(self, event: Event, state: State) -> bool:
        """Match on AFTER_DATALOADER event, but only if parametrized to do so. """
        is_event = (event == Event.AFTER_DATALOADER)
        if not is_event:
            return False

        is_keep = (self.hparams.keep < 1)
        if not is_keep:
            return False

        is_chosen = do_selective_backprop(
            epoch=state.epoch,
            batch_idx=state.batch_idx,
            start_epoch=int(state.max_epochs * self.hparams.start),
            end_epoch=int(state.max_epochs * self.hparams.end),
            interrupt=self.hparams.interrupt,
        )
        return is_chosen

    def apply(self, event: Event, state: State, logger: Optional[Logger] = None) -> None:
        """Apply selective backprop to the current batch.
        """
        input, target = state.batch_pair
        assert isinstance(input, Tensor) and isinstance(target, Tensor), \
            "Multiple tensors not supported for this method yet."

        assert callable(state.model.module.loss)  # type: ignore - type not found

        # Model expected to only take in input, not the full batch
        model = lambda X: state.model((X, None))

        def loss(p, y, reduction="none"):
            return state.model.module.loss(p, (None, y), reduction=reduction)  # type: ignore

        with state.precision_context(state.precision):
            new_input, new_target = selective_backprop(
                input, target, model, loss, self.hparams.keep,
                self.hparams.scale_factor)  # type: ignore - ditto because of loss
        state.batch = (new_input, new_target)



from __future__ import annotations

import atexit
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Optional

import torch.profiler
import yahp as hp
from torch.profiler.profiler import ProfilerAction

from composer import Callback
from composer.callbacks.callback_hparams import CallbackHparams
from composer.core.types import StateDict
from composer.utils.ddp import get_global_rank

if TYPE_CHECKING:
    from composer.core import Logger, State

_PROFILE_MISSING_ERROR = "The profiler has not been setup. Please call profiler.training_start() before training starts."


@dataclass
class TorchProfilerHparams(CallbackHparams):
    tensorboard_trace_handler_dir: str = hp.required("directory to store trace results")
    tensorboard_use_gzip: bool = hp.optional("use gzip for trace", default=False)
    record_shapes: bool = hp.optional(doc="record_shapes", default=True)
    profile_memory: bool = hp.optional(doc="track tensor memory allocations and frees", default=False)
    with_stack: bool = hp.optional(doc="record stack info", default=True)
    with_flops: bool = hp.optional(doc="estimate flops for operators", default=True)

    skip: int = hp.optional("Number of batches to skip at epoch start", default=0)
    warmup: int = hp.optional("Number of warmup batches in a cycle", default=1)
    active: int = hp.optional("Number of batches to profile in a cycle", default=5)
    wait: int = hp.optional("Number of batches to skip at the end of each cycle", default=0)

    def initialize_object(self) -> TorchProfiler:
        return TorchProfiler(**asdict(self))


@dataclass
class TorchProfilerState:
    batch_in_epoch: int = 0
    batches_per_epoch: int = 0


class TorchProfiler(Callback):

    def __init__(
        self,
        *,
        skip: int,
        warmup: int,
        active: int,
        wait: int,
        tensorboard_trace_handler_dir: str,
        tensorboard_use_gzip: bool,
        record_shapes: bool,
        profile_memory: bool,
        with_stack: bool,
        with_flops: bool,
    ) -> None:
        super().__init__()
        self.hparams = TorchProfilerHparams(
            skip=skip,
            warmup=warmup,
            active=active,
            wait=wait,
            tensorboard_trace_handler_dir=tensorboard_trace_handler_dir,
            tensorboard_use_gzip=tensorboard_use_gzip,
            record_shapes=record_shapes,
            profile_memory=profile_memory,
            with_stack=with_stack,
            with_flops=with_flops,
        )
        self.profiler: Optional[torch.profiler.profile] = None
        self.profiler_state: TorchProfilerState = TorchProfilerState()
        self._torch_profiler_scheduler = torch.profiler.profiler.schedule(
            wait=self.hparams.wait,
            warmup=self.hparams.warmup,
            active=self.hparams.active,
            skip_first=self.hparams.skip,
        )

    def state_dict(self) -> StateDict:
        return asdict(self.profiler_state)

    def load_state_dict(self, state: StateDict) -> None:
        self.profiler_state = TorchProfilerState(**state)

    def scheduler_fn(self, profiler_step: int) -> ProfilerAction:
        # Invoked on every batch, at the batch end
        # But, it's called one batch in advance.
        # Wrapping the default scheduling function to deal with epoch boundaries
        # Giving the torch scheduler the batch in the epoch, not the global step

        # adding 1 since this is called before the step is incremented
        next_batch_in_epoch = self.profiler_state.batch_in_epoch + 1
        if profiler_step == 0:
            next_batch_in_epoch = 0
        torch_scheduler_action = self._torch_profiler_scheduler(next_batch_in_epoch)
        if next_batch_in_epoch == self.profiler_state.batches_per_epoch:
            if torch_scheduler_action == ProfilerAction.RECORD:
                # force saving at epoch boundaries
                torch_scheduler_action = ProfilerAction.RECORD_AND_SAVE
        return torch_scheduler_action

    def training_start(self, state: State, logger: Logger) -> None:
        assert self.profiler is None, _PROFILE_MISSING_ERROR
        self.profiler = torch.profiler.profile(
            schedule=self.scheduler_fn,
            on_trace_ready=torch.profiler.tensorboard_trace_handler(
                self.hparams.tensorboard_trace_handler_dir,
                worker_name=str(get_global_rank()),
                use_gzip=self.hparams.tensorboard_use_gzip,
            ),
            activities=None,  # auto-set
            record_shapes=self.hparams.record_shapes,
            profile_memory=self.hparams.profile_memory,
            with_stack=self.hparams.with_stack,
            with_flops=self.hparams.with_flops,
        )
        self.profiler.__enter__()
        atexit.register(self._close_profiler)

    def batch_end(self, state: State, logger: Logger) -> None:
        assert self.profiler is not None, _PROFILE_MISSING_ERROR
        self.profiler.step()

    def epoch_start(self, state: State, logger: Logger) -> None:
        self.profiler_state.batches_per_epoch = state.steps_per_epoch

    def batch_start(self, state: State, logger: Logger) -> None:
        self.profiler_state.batch_in_epoch = state.batch_idx
        assert self.profiler is not None, _PROFILE_MISSING_ERROR
        logger.metric_batch({"profiler/state": self.profiler.current_action.name})

    def _close_profiler(self) -> None:
        assert self.profiler is not None
        self.profiler.__exit__(None, None, None)



# Copyright 2021 MosaicML. All Rights Reserved.

import functools
import os
import random
import shutil
from logging import Logger
from typing import Dict, Optional

import pytest
import torch
from _pytest.monkeypatch import MonkeyPatch

from composer.callbacks.callback_hparams import CallbackHparams
from composer.core.callback import Callback
from composer.core.event import Event
from composer.core.state import State
from composer.core.types import StateDict
from composer.trainer.ddp import FileStoreHparams
from composer.trainer.devices import CPUDeviceHparams, DeviceHparams, GPUDeviceHparams
from composer.trainer.trainer import Trainer
from composer.trainer.trainer_hparams import TrainerHparams, callback_registry
from tests.test_state import assert_state_equivalent
from tests.utils.deep_compare import deep_compare


class DummyStatefulCallback(Callback):

    def __init__(self) -> None:
        super().__init__()
        self.random_value = random.random()

    def state_dict(self) -> StateDict:
        return {
            "random_value": self.random_value,
        }

    def load_state_dict(self, state: StateDict) -> None:
        self.random_value = state["random_value"]


class DummyStatefulCallbackHparams(CallbackHparams):

    def initialize_object(self) -> DummyStatefulCallback:
        return DummyStatefulCallback()


class EventCounterCallback(Callback):

    def __init__(self) -> None:
        self.event_to_num_calls: Dict[Event, int] = {}

        for event in Event:
            self.event_to_num_calls[event] = 0
            setattr(self, event.value, functools.partial(self._event_catchall, event=event))

    def _event_catchall(self, state: State, logger: Logger, event: Event):
        if event == Event.TRAINING_START:
            # ignoring training start as it is called once per startup
            # and the states otherwise won't match
            return
        self.event_to_num_calls[event] += 1

    def state_dict(self) -> StateDict:
        return {"events": self.event_to_num_calls}

    def load_state_dict(self, state: StateDict) -> None:
        self.event_to_num_calls.update(state["events"])


class EventCounterCallbackHparams(CallbackHparams):

    def initialize_object(self) -> EventCounterCallback:
        return EventCounterCallback()


@pytest.fixture
def checkpoint_folder(ddp_tmpdir) -> str:
    return os.path.join(ddp_tmpdir, "checkpoints")


@pytest.fixture
def checkpointing_trainer_hparams(mosaic_trainer_hparams: TrainerHparams, ddp_tmpdir: str) -> TrainerHparams:
    mosaic_trainer_hparams.grad_accum = 2
    mosaic_trainer_hparams.checkpoint_interval = 1
    mosaic_trainer_hparams.checkpoint_folder = os.path.join(ddp_tmpdir, "checkpoints")
    mosaic_trainer_hparams.max_epochs = 2
    mosaic_trainer_hparams.ddp.fork_rank_0 = False
    mosaic_trainer_hparams.callbacks.append(DummyStatefulCallbackHparams())
    mosaic_trainer_hparams.callbacks.append(EventCounterCallbackHparams())
    return mosaic_trainer_hparams


is_main_pytest_process: Optional[bool] = None


def _should_create_trainer(trainer_key: str) -> bool:
    # trainer_key is used for DDP forking, where a single test may create multiple trainers
    # each trainer should have a unique key. This key is stored in an environ, so when the forked process is created,
    # the test knows whether it got to the same point in the test as the main process.
    #
    # if the trainer key differs -- i.e. the test is forking for the 2nd trainer in a test, but the forked process
    # just got to the first trainer, or a forked process from the first trainer got to the point where it was going
    # to create the 2nd trainer, then _should_create_trainer returns False
    # otherwise, it returns True
    global is_main_pytest_process
    if is_main_pytest_process is None:
        is_main_pytest_process = "TRAINER_KEY" not in os.environ

    if is_main_pytest_process:
        os.environ["TRAINER_KEY"] = trainer_key
        os.environ.pop("RANK", None)
        os.environ.pop("LOCAL_RANK", None)
        os.environ.pop("WORLD_SIZE", None)
    return os.environ["TRAINER_KEY"] == trainer_key


def get_trainer(device_hparams: DeviceHparams,
                trainer_hparams: TrainerHparams,
                *,
                trainer_key: str,
                ddp_tmpdir: str,
                checkpoint_filepath: Optional[str] = None) -> Optional[Trainer]:
    trainer_hparams.device = device_hparams
    trainer_hparams.ddp.store = FileStoreHparams(file_name=os.path.join(ddp_tmpdir, f"store_{trainer_key}"))
    trainer_hparams.checkpoint_filepath = checkpoint_filepath
    if not _should_create_trainer(trainer_key):
        return
    return Trainer.create_from_hparams(hparams=trainer_hparams)


def move_checkpoint(checkpoint_folder: str, name: str, destination_path: str) -> None:
    global is_main_pytest_process
    if not is_main_pytest_process:
        return
    checkpoint_filepath = os.path.join(checkpoint_folder, name)
    os.rename(checkpoint_filepath, destination_path)


def move_hparams(checkpoint_folder: str, destination_path: str) -> None:
    global is_main_pytest_process
    if not is_main_pytest_process:
        return
    checkpoint_filepath = os.path.join(checkpoint_folder, "hparams.yaml")
    os.rename(checkpoint_filepath, destination_path)


def assert_checkpoints_equivalent(hparams_file_a: str, checkpoint_file_a: str, hparams_file_b: str,
                                  checkpoint_file_b: str) -> None:
    checkpoint_a = torch.load(checkpoint_file_a, map_location='cpu')
    checkpoint_b = torch.load(checkpoint_file_b, map_location='cpu')

    hparams_a = TrainerHparams.create(hparams_file_a, cli_args=False)
    assert isinstance(hparams_a, TrainerHparams)
    hparams_b = TrainerHparams.create(hparams_file_b, cli_args=False)
    assert isinstance(hparams_b, TrainerHparams)

    # manually fix the store, as that part of the hparams WILL differ
    hparams_b.ddp.store = hparams_a.ddp.store

    hparams_a.checkpoint_filepath = hparams_b.checkpoint_filepath

    assert hparams_a.to_dict() == hparams_b.to_dict()

    hparams_a.checkpoint_filepath = checkpoint_file_a
    hparams_b.checkpoint_filepath = checkpoint_file_b

    trainer_a = Trainer.create_from_hparams(hparams=hparams_a)
    trainer_b = Trainer.create_from_hparams(hparams=hparams_b)

    state_a = trainer_a.state
    state_b = trainer_b.state
    assert_state_equivalent(state_a, state_b, skip_transient_fields=True)

    deep_compare(checkpoint_a["rng"], checkpoint_b["rng"])


@pytest.fixture(autouse=True)
def inject_stateful_callback_hparams(monkeypatch: MonkeyPatch):
    monkeypatch.setitem(callback_registry, "dummy", DummyStatefulCallbackHparams)
    monkeypatch.setitem(callback_registry, "event_counter", EventCounterCallbackHparams)


def clear_checkpoint_folder(checkpoint_folder: str):
    global is_main_pytest_process
    if not is_main_pytest_process:
        return
    shutil.rmtree(checkpoint_folder, ignore_errors=True)


@pytest.mark.run_long
@pytest.mark.timeout(90)
@pytest.mark.parametrize("device_hparams", [
    pytest.param(CPUDeviceHparams(n_cpus=1), id="1cpu"),
    pytest.param(CPUDeviceHparams(n_cpus=2), id='2cpu'),
    pytest.param(GPUDeviceHparams(n_gpus=1), marks=pytest.mark.n_gpus(1), id="1gpu"),
    pytest.param(GPUDeviceHparams(n_gpus=2), marks=pytest.mark.n_gpus(2), id="2gpu"),
])
@pytest.mark.parametrize("checkpoint_filename", ["ep1", "it4", "it1", "it6"])
@pytest.mark.parametrize("validate_every_n_batches,validate_every_n_epochs", [
    (0, 1),
    (1, 0),
])
def test_checkpoint(
    device_hparams: DeviceHparams,
    checkpointing_trainer_hparams: TrainerHparams,
    ddp_tmpdir: str,
    checkpoint_folder: str,
    checkpoint_filename: str,
    validate_every_n_batches: int,
    validate_every_n_epochs: int,
):
    """strategy:
    - train two epochs. capture checkpoints after `checkpoint_interval` and ep2.
    - create a new trainer from the `checkpoint_interval` checkpoint, and train until end. checkpoint again.
    - assert that the checkpoint from the new trainer at the end is the same as the checkpoint from the first trainer at the end.
    """
    checkpointing_trainer_hparams.checkpoint_interval_unit = "ep" if checkpoint_filename.startswith("ep") else "it"
    checkpointing_trainer_hparams.validate_every_n_batches = validate_every_n_batches
    checkpointing_trainer_hparams.validate_every_n_epochs = validate_every_n_epochs
    first_trainer = get_trainer(
        device_hparams=device_hparams,
        trainer_hparams=checkpointing_trainer_hparams,
        trainer_key="first",
        ddp_tmpdir=ddp_tmpdir,
    )
    checkpoint_a_file_path = os.path.join(ddp_tmpdir, "checkpoint_a.pt")
    checkpoint_b_file_path = os.path.join(ddp_tmpdir, 'checkpoint_b.pt')
    trainer_1_hparams_filepath = os.path.join(ddp_tmpdir, "trainer_1_hparams.yaml")
    final_checkpoint = "ep2.pt" if checkpoint_filename.startswith("ep") else "it8.pt"
    if first_trainer is not None:
        first_trainer.fit()
        move_checkpoint(checkpoint_folder, f"{checkpoint_filename}.pt", checkpoint_a_file_path)
        move_checkpoint(checkpoint_folder, final_checkpoint, checkpoint_b_file_path)
        move_hparams(checkpoint_folder, trainer_1_hparams_filepath)
        clear_checkpoint_folder(checkpoint_folder)
        validate_events_called_expected_number_of_times(first_trainer)
    second_trainer_hparams = TrainerHparams.create(trainer_1_hparams_filepath, cli_args=False)
    assert isinstance(second_trainer_hparams, TrainerHparams)
    second_trainer = get_trainer(
        device_hparams=device_hparams,
        trainer_hparams=second_trainer_hparams,
        trainer_key="second",
        ddp_tmpdir=ddp_tmpdir,
        checkpoint_filepath=checkpoint_a_file_path,
    )
    checkpoint_c_file_path = os.path.join(ddp_tmpdir, "checkpoint_c.pt")
    trainer_2_hparams_filepath = os.path.join(ddp_tmpdir, "trainer_2_hparams.yaml")
    if second_trainer is not None:
        second_trainer.fit()
        move_checkpoint(checkpoint_folder, final_checkpoint, checkpoint_c_file_path)
        move_hparams(checkpoint_folder, trainer_2_hparams_filepath)
        clear_checkpoint_folder(checkpoint_folder)
        validate_events_called_expected_number_of_times(second_trainer)

    if is_main_pytest_process:
        assert_checkpoints_equivalent(
            hparams_file_a=trainer_1_hparams_filepath,
            checkpoint_file_a=checkpoint_b_file_path,
            hparams_file_b=trainer_2_hparams_filepath,
            checkpoint_file_b=checkpoint_c_file_path,
        )


def validate_events_called_expected_number_of_times(trainer: Trainer):
    state = trainer.state

    num_epochs = state.max_epochs
    num_total_steps = num_epochs * state.steps_per_epoch
    num_total_microbatches = num_total_steps * state.grad_accum
    num_evals = 0
    if trainer.validate_every_n_batches > 0:
        num_evals = num_total_steps // trainer.validate_every_n_batches
    if trainer.validate_every_n_epochs > 0:
        num_evals = num_epochs // trainer.validate_every_n_epochs
    assert state.eval_dataloader is not None
    num_eval_steps = num_evals * len(state.eval_dataloader)

    event_to_num_expected_invocations = {
        Event.INIT: 1,
        # training start is being ignored, as it should be called once per startup
        Event.TRAINING_START: 0,
        Event.EPOCH_START: num_epochs,
        Event.BATCH_START: num_total_steps,
        Event.AFTER_DATALOADER: num_total_steps,
        Event.BEFORE_FORWARD: num_total_microbatches,
        Event.AFTER_FORWARD: num_total_microbatches,
        Event.BEFORE_LOSS: num_total_microbatches,
        Event.AFTER_LOSS: num_total_microbatches,
        Event.BEFORE_BACKWARD: num_total_microbatches,
        Event.AFTER_BACKWARD: num_total_microbatches,
        Event.BEFORE_TRAIN_BATCH: num_total_steps,
        Event.AFTER_TRAIN_BATCH: num_total_steps,
        Event.BATCH_END: num_total_steps,
        Event.EPOCH_END: num_epochs,
        Event.EVAL_START: num_evals,
        Event.EVAL_BATCH_START: num_eval_steps,
        Event.EVAL_BEFORE_FORWARD: num_eval_steps,
        Event.EVAL_AFTER_FORWARD: num_eval_steps,
        Event.EVAL_BATCH_END: num_eval_steps,
        Event.EVAL_END: num_evals,
        Event.TRAINING_END: 1,
    }

    for callback in trainer.engine.callbacks:
        if isinstance(callback, EventCounterCallback):
            for event, expected in event_to_num_expected_invocations.items():
                actual = callback.event_to_num_calls[event]
                assert expected == actual, f"Event {event} expected to be called {expected} times, but instead it was called {actual} times"
            return
    assert False, "EventCounterCallback not found in callbacks"



# Copyright 2021 MosaicML. All Rights Reserved.

import warnings

import torch.distributed as dist


def get_global_rank() -> int:
    if not dist.is_available():
        warnings.warn("Torch distributed is not available; returning 0 for global rank")
        return 0
    return dist.get_rank()  # raises a runtime error if not yet initialized


def is_rank_zero() -> bool:
    return get_global_rank() == 0


def is_rank_set() -> bool:
    try:
        get_global_rank()
        return True
    except RuntimeError:
        return False



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.channels_last.channels_last import ChannelsLast as ChannelsLast
from composer.algorithms.channels_last.channels_last import ChannelsLastHparams as ChannelsLastHparams

_name = 'Channels Last'
_class_name = 'ChannelsLast'
_functional = ''
_tldr = 'Uses channels last memory format (NHWC)'
_attribution = 'PyTorch'
_link = 'https://pytorch.org/tutorials/intermediate/memory_format_tutorial.html'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import os

import pytest

import composer
from composer.trainer import TrainerHparams


def walk_model_yamls():
    yamls = []
    for root, dirs, files in os.walk(os.path.join(os.path.dirname(composer.__file__), "yamls", "models")):
        for name in files:
            filepath = os.path.join(root, name)
            if filepath.endswith(".yaml"):
                yamls.append(filepath)
    assert len(yamls) > 0, "there should be at least one yaml!"
    return yamls


@pytest.mark.parametrize("hparams_file", walk_model_yamls())
class TestHparamsCreate:

    def test_hparams_create(self, hparams_file: str):
        hparams = TrainerHparams.create(hparams_file)
        assert isinstance(hparams, TrainerHparams)


# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from composer.core.serializable import Serializable

if TYPE_CHECKING:
    from composer.core import Event, Logger, State


class Algorithm(Serializable, ABC):
    """Base class for algorithms.

    Algorithms are pieces of code which run at specific events in the training loop. Algorithms
    modify the trainer's state, generally with the effect of improving the model's quality, or
    increasing the efficiency and throughput of the training loop. 
    
    Algorithms must implement two methods:
    :func:`match`, which returns whether the algorithm should be run given
    the current event and state, and :func:`apply`, which makes an in-place
    change to the State.
    """

    @property
    def find_unused_parameters(self) -> bool:
        """Indicates that the effect of this algorithm may cause some model parameters to be unused.

        Used to tell DDP that some parameters will be frozen during
        training and hence it should not expect gradients from them.
        All algorithms which do any kind of parameter freezing should
        override this function to return True.
        """
        return False

    @abstractmethod
    def match(self, event: Event, state: State) -> bool:
        """Determines whether this algorithm should run, given the current :class:`Event` and :class:`State`.

        Examples:
        
        To only run on a specific event:
            
            >>> return event == Event.BEFORE_LOSS

        Switching based on state attributes:
            
            >>> return state.epoch > 30 && state.world_size == 1

        See :class:`State` for accessible attributes.

        Args:
            event (:class:`Event`): The current event.
            state (:class:`State`): The current state.
        Returns:
            bool: True if this algorithm should run now.
        """
        raise NotImplementedError(f'implement match() required for {self.__class__.__name__}')

    @abstractmethod
    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """Applies the algorithm to make an in-place change to the State
        
        Can optionally return an exit code to be stored in a :class:`Trace`.

        Args:
            event (:class:`Event`): The current event.
            state (:class:`State`): The current state.
            logger (:class:`Logger`): A logger to use for logging algorithm-specific metrics.
        Returns:
            int or None: exit code that is stored in :class:`Trace` and made accessible for debugging.
        """
        raise NotImplementedError(f'implement apply() required for {self.__class__.__name__}')

    def __str__(self) -> str:
        """Returns the class name."""
        return self.__class__.__name__



# Copyright 2021 MosaicML. All Rights Reserved.

# Code adapted from (Generic) EfficientNets for PyTorch repo:
# https://github.com/rwightman/gen-efficientnet-pytorch
import math
import re
from typing import Any, Callable, Dict

import torch
import torch.nn as nn


def round_channels(
    channels: float,
    width_multiplier: float,
    divisor: int = 8,
    min_value: int = None,
) -> int:
    """Round number of channels after scaling with width multiplier. This 
       function ensures that channel integers halfway inbetween divisors 
       is rounded up.

    Args:
        channels (float): Number to round.
        width_multiplier (float): Amount to scale `channels`.
        divisor (int): Number to make the output divisible by.
        min_value (int): Minimum value the output can be.
    """

    if not width_multiplier:
        return int(channels)
    channels *= width_multiplier

    min_value = min_value or divisor
    new_channels = max(min_value, int(channels + divisor / 2) // divisor * divisor)
    if new_channels < 0.9 * channels:  # increase channels if rounding decreases by >10%
        new_channels += divisor
    return new_channels


def calculate_same_padding(kernel_size, dilation, stride):
    """Calculates the amount of padding to use to get the "SAME" functionality in Tensorflow.
    """

    return ((stride - 1) + dilation * (kernel_size - 1)) // 2


def drop_connect(inputs: torch.Tensor, drop_connect_rate: float, training: bool):
    """ Randomly mask a set of samples. Provides similar regularization as stochastic depth.
    Args:
        input (torch.Tensor): Input tensor to mask.
        drop_connect_rate (float): Probability of droppping each sample.
        training (bool): Whether or not the model is training
    """

    if not training:
        return inputs

    keep_prob = 1 - drop_connect_rate
    rand_tensor = keep_prob + torch.rand(
        [inputs.size()[0], 1, 1, 1],
        dtype=inputs.dtype,
        device=inputs.device,
    )
    rand_tensor.floor_()  # binarize
    output = inputs.div(keep_prob) * rand_tensor
    return output


class SqueezeExcite(nn.Module):
    """Squeeze Excite Layer
    
    Args:
        in_channels (int): Number of channels in the input tensor.
        latent_channels (int): Number of hidden channels.
        act_layer (torch.nn.Module): Activation layer to use in block.
    """

    def __init__(
        self,
        in_channels: int,
        latent_channels: int,
        act_layer: Callable[..., nn.Module] = nn.ReLU,
    ):
        super().__init__()

        self.global_avg_pool = nn.AdaptiveAvgPool2d(1)
        self.conv_reduce = nn.Conv2d(in_channels, latent_channels, kernel_size=1, bias=True)
        self.act1 = act_layer(inplace=True)
        self.conv_expand = nn.Conv2d(latent_channels, in_channels, kernel_size=1, bias=True)
        self.gate_fn = torch.nn.Sigmoid()

    def forward(self, x: torch.Tensor):
        out = self.global_avg_pool(x)
        out = self.conv_reduce(out)
        out = self.act1(out)
        out = self.conv_expand(out)
        out = x * self.gate_fn(out)
        return out


class DepthwiseSeparableConv(nn.Module):
    """Depthwise Separable Convolution layer
    
    Args:
        in_channels (int): Number of channels in the input tensor.
        out_channels (int): Number of channels in the output tensor.
        kernel_size (int): Size of the convolving kernel.
        stride (int): Stride of the convolution.
        se_ratio (float): How much to scale `in_channels` for the hidden layer 
                          dimensionality of the squeeze-excite module.
        drop_connect_rate (float): Probability of dropping a sample before the
                                   identity connection, provides regularization
                                   similar to stochastic depth.
        act_layer (torch.nn.Module): Activation layer to use in block.
        norm_kwargs (dict): Normalization layer's keyword arguments.
        norm_layer (torch.nn.Module): Normalization layer to use in block.    
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 stride: int,
                 se_ratio: float,
                 drop_connect_rate: float,
                 act_layer: Callable[..., nn.Module],
                 norm_kwargs: dict,
                 norm_layer: Callable[..., nn.Module] = nn.BatchNorm2d):
        super().__init__()
        self.drop_connect_rate = drop_connect_rate
        self.has_residual = (in_channels == out_channels and stride == 1)
        self.has_se = se_ratio is not None and se_ratio > 0.0

        padding = calculate_same_padding(kernel_size, dilation=1, stride=stride)
        self.conv_depthwise = nn.Conv2d(in_channels=in_channels,
                                        out_channels=in_channels,
                                        groups=in_channels,
                                        kernel_size=kernel_size,
                                        stride=stride,
                                        padding=padding,
                                        bias=False)
        self.bn1 = norm_layer(in_channels, **norm_kwargs)
        self.act1 = act_layer(inplace=True)

        if self.has_se:
            latent_channels = max(1, int(in_channels * se_ratio))
            self.se = SqueezeExcite(in_channels, latent_channels, act_layer)

        self.conv_pointwise = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=1,
            bias=False,
        )
        self.bn2 = norm_layer(out_channels, **norm_kwargs)
        self.act2 = act_layer(inplace=True)

    def forward(self, input: torch.Tensor):
        residual = input

        out = self.conv_depthwise(input)
        out = self.bn1(out)
        out = self.act1(out)

        if self.has_se:
            out = self.se(out)

        out = self.conv_pointwise(out)
        out = self.bn2(out)
        out = self.act2(out)

        if self.has_residual:
            if self.drop_connect_rate > 0.0:
                out = drop_connect(out, self.drop_connect_rate, self.training)
            out += residual
        return out


class MBConvBlock(nn.Module):
    """Mobile Inverted Residual Bottleneck Block as defined in 
       https://arxiv.org/abs/1801.04381
    
    Args:
        in_channels (int): Number of channels in the input tensor.
        out_channels (int): Number of channels in the output tensor.
        kernel_size (int): Size of the convolving kernel.
        stride (int): Stride of the convolution.
        expand_ratio (int): How much to expand the input channels for the
                            depthwise convolution.
        se_ratio (float): How much to scale `in_channels` for the hidden layer
                          dimensionality of the squeeze-excite module.
        drop_connect_rate (float): Probability of dropping a sample before the
                                   identity connection, provides regularization
                                   similar to stochastic depth.
        act_layer (torch.nn.Module): Activation layer to use in block.
        norm_kwargs (dict): Normalization layer's keyword arguments.
        norm_layer (torch.nn.Module): Normalization layer to use in block.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: int,
                 stride: int,
                 expand_ratio: int,
                 se_ratio: float,
                 drop_connect_rate: float,
                 act_layer: Callable[..., nn.Module],
                 norm_kwargs: dict,
                 norm_layer: Callable[..., nn.Module] = nn.BatchNorm2d):
        super().__init__()
        self.drop_connect_rate = drop_connect_rate
        self.has_residual = (in_channels == out_channels and stride == 1)
        self.has_se = se_ratio is not None and se_ratio > 0.0

        mid_channels = round_channels(in_channels, expand_ratio)

        # Point-wise convolution expansion
        self.conv1x1_expand = nn.Conv2d(in_channels, mid_channels, kernel_size=1, bias=False)
        self.bn1 = norm_layer(mid_channels, **norm_kwargs)
        self.act1 = act_layer(inplace=True)

        # Depth-wise Convolution
        padding = calculate_same_padding(kernel_size, dilation=1, stride=stride)
        self.conv_depthwise = nn.Conv2d(in_channels=mid_channels,
                                        out_channels=mid_channels,
                                        groups=mid_channels,
                                        kernel_size=kernel_size,
                                        stride=stride,
                                        padding=padding,
                                        bias=False)
        self.bn2 = norm_layer(mid_channels, **norm_kwargs)
        self.act2 = act_layer(inplace=True)

        # Squeeze and Excitation layer, if specified
        if self.has_se:
            latent_channels = max(1, int(in_channels * se_ratio))
            self.se = SqueezeExcite(mid_channels, latent_channels, act_layer)

        # Point-wise convolution contraction
        self.conv1x1_contract = nn.Conv2d(mid_channels, out_channels, kernel_size=1, bias=False)
        self.bn3 = norm_layer(out_channels, **norm_kwargs)

    def forward(self, input: torch.Tensor):
        residual = input

        out = self.conv1x1_expand(input)
        out = self.bn1(out)
        out = self.act1(out)

        out = self.conv_depthwise(out)
        out = self.bn2(out)
        out = self.act2(out)

        if self.has_se:
            out = self.se(out)

        out = self.conv1x1_contract(out)
        out = self.bn3(out)

        if self.has_residual:
            if self.drop_connect_rate:
                out = drop_connect(out, self.drop_connect_rate, self.training)
            out += residual
        return out


class EfficientNet(nn.Module):
    """EfficientNet architecture designed for ImageNet in 
       https://arxiv.org/abs/1905.11946

       Args:
           num_classes (int): Size of the EfficientNet output, typically viewed
                              as the number of classes in a classification task.
           width_multiplier (float): How much to scale the EfficientNet-B0 channel
                                     dimension throughout the model. 
           depth_multiplier (float): How much to scale the EFficientNet-B0 depth.
           drop_rate (float): Dropout probability for the penultimate activations.
           drop_connect_rate (float): Probability of dropping a sample before the
                                      identity connection, provides regularization
                                      similar to stochastic depth.
           act_layer (torch.nn.Module): Activation layer to use in the model.
           norm_kwargs (dict): Normalization layer's keyword arguments.
           norm_layer (torch.nn.Module): Normalization layer to use in the model.
    """

    # EfficientNet-B0 architecture specification
    _blocks_strings = [
        'r1_k3_s1_e1_i32_o16_se0.25',
        'r2_k3_s2_e6_i16_o24_se0.25',
        'r2_k5_s2_e6_i24_o40_se0.25',
        'r3_k3_s2_e6_i40_o80_se0.25',
        'r3_k5_s1_e6_i80_o112_se0.25',
        'r4_k5_s2_e6_i112_o192_se0.25',
        'r1_k3_s1_e6_i192_o320_se0.25',
    ]

    def __init__(self,
                 num_classes: int,
                 width_multiplier: float = 1.0,
                 depth_multiplier: float = 1.0,
                 drop_rate: float = 0.2,
                 drop_connect_rate: float = 0.2,
                 act_layer: Callable[..., nn.Module] = nn.SiLU,
                 norm_kwargs: dict = {
                     "momentum": 0.1,
                     "eps": 1e-5
                 },
                 norm_layer: Callable[..., nn.Module] = nn.BatchNorm2d):
        super(EfficientNet, self).__init__()
        self.num_classes = num_classes

        in_channels = 3
        out_channels = round_channels(32, width_multiplier)
        padding = calculate_same_padding(kernel_size=3, dilation=1, stride=2)
        self.conv_stem = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=2,
            padding=padding,
            bias=False,
        )
        self.bn1 = norm_layer(num_features=out_channels, **norm_kwargs)
        self.act1 = act_layer(inplace=True)

        # Count the number of blocks in the model
        block_count = 0.
        for block_string in self._blocks_strings:
            _, num_repeat = self._decode_block_string(block_string)
            block_count += num_repeat

        # Decode block strings and add blocks
        block_idx = 0.
        blocks = []
        block_args = {}
        for block_string in self._blocks_strings:
            block_args, num_repeat = self._decode_block_string(block_string)
            # Scale channels and number of repeated blocks based on multipliers
            block_args["in_channels"] = round_channels(
                block_args["in_channels"],
                width_multiplier,
            )
            block_args["out_channels"] = round_channels(
                block_args["out_channels"],
                width_multiplier,
            )
            num_repeat = int(math.ceil(depth_multiplier * num_repeat))

            # Add activation, normalization layers, and drop connect
            block_args['act_layer'] = act_layer
            block_args['norm_kwargs'] = norm_kwargs
            block_args['norm_layer'] = norm_layer

            # Delete expand_ratio when set to 1 to use depthwise separable convolution layer
            if block_args['expand_ratio'] == 1:
                del block_args['expand_ratio']

            for i in range(num_repeat):
                # Linearly decay drop_connect_rate across model depth
                block_args['drop_connect_rate'] = drop_connect_rate * block_idx / block_count

                if 'expand_ratio' not in block_args:
                    blocks.append(DepthwiseSeparableConv(**block_args))
                else:
                    blocks.append(MBConvBlock(**block_args))
                block_idx += 1

                # Only the first block in a stage can have stride != 1
                if i == 0:
                    block_args['stride'] = 1
                    block_args['in_channels'] = block_args['out_channels']

        self.blocks = nn.Sequential(*blocks)

        in_channels = block_args['out_channels']
        out_channels = round_channels(1280, width_multiplier)
        self.conv_head = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self.bn2 = norm_layer(out_channels, **norm_kwargs)
        self.act2 = act_layer(inplace=True)

        self.global_avg_pool = nn.AdaptiveAvgPool2d(1)
        self.dropout = nn.Dropout(drop_rate)
        self.classifier = nn.Linear(out_channels, num_classes)

        # Initialization from gen-efficientnet-pytorch repo
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                fan_out = (m.kernel_size[0] * m.kernel_size[1] * m.out_channels) // m.groups
                m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, torch.nn.BatchNorm2d):
                m.weight.data.fill_(1.0)
                m.bias.data.zero_()
            elif isinstance(m, torch.nn.Linear):
                fan_out = m.weight.size(0)
                init_range = 1.0 / math.sqrt(fan_out)
                m.weight.data.uniform_(-init_range, init_range)
                m.bias.data.zero_()

    def extract_features(self, input: torch.Tensor):
        out = self.conv_stem(input)
        out = self.bn1(out)
        out = self.act1(out)
        out = self.blocks(out)
        out = self.conv_head(out)
        out = self.bn2(out)
        out = self.act2(out)
        out = self.global_avg_pool(out)
        return out.flatten(1)

    def forward(self, input: torch.Tensor):
        out = self.extract_features(input)
        out = self.dropout(out)
        return self.classifier(out)

    @staticmethod
    def get_model_from_name(model_name: str, num_classes, drop_connect_rate: float):
        """ Instantiate an EfficientNet model family member based on the model_name string
        """

        # Coefficients: width, depth, res, dropout
        model_arch = {
            'efficientnet-b0': (1.0, 1.0, 224, 0.2),
            'efficientnet-b1': (1.0, 1.1, 240, 0.2),
            'efficientnet-b2': (1.1, 1.2, 260, 0.3),
            'efficientnet-b3': (1.2, 1.4, 300, 0.3),
            'efficientnet-b4': (1.4, 1.8, 380, 0.4),
            'efficientnet-b5': (1.6, 2.2, 456, 0.4),
            'efficientnet-b6': (1.8, 2.6, 528, 0.5),
            'efficientnet-b7': (2.0, 3.1, 600, 0.5),
        }

        model_params = model_arch[model_name]
        width_multiplier = model_params[0]
        depth_multiplier = model_params[1]
        drop_rate = model_params[3]
        return EfficientNet(num_classes=num_classes,
                            width_multiplier=width_multiplier,
                            depth_multiplier=depth_multiplier,
                            drop_rate=drop_rate,
                            drop_connect_rate=drop_connect_rate)

    def _decode_block_string(self, block_string: str):
        """ Decodes an EfficientNet block specification string into a dictionary
            of keyword arguments for a block in the architecture
        """
        arg_strings = block_string.split('_')
        args = {}
        for arg_string in arg_strings:
            splits = re.split(r'(\d.*)', arg_string)
            if len(splits) >= 2:
                key, value = splits[:2]
                args[key] = value
        num_repeat = int(args['r'])
        block_args = dict(kernel_size=int(args['k']),
                          stride=int(args['s']),
                          expand_ratio=int(args['e']),
                          in_channels=int(args['i']),
                          out_channels=int(args['o']),
                          se_ratio=float(args['se']) if 'se' in args else None)  # type: Dict[str, Any]
        return block_args, num_repeat



# Copyright 2021 MosaicML. All Rights Reserved.

"""
Helper function to generate the README table

"""
from pathlib import Path

from composer import functional as CF
from composer import utils

HEADER = ['Name', 'Functional', 'Attribution', 'tl;dr', 'Card']
ATTRIBUTES = ['_class_name', '_functional', '_tldr', '_attribution', '_link', '_method_card']
GITHUB_BASE = 'https://github.com/mosaicml/mosaicml/tree/dev/composer/algorithms/'  # TODO: update for launch

folder_path = Path(__file__).parent
methods = utils.list_dirs(folder_path)

if not len(methods):
    raise ValueError(f'Found 0 methods in {folder_path}')

print(f'Found {len(methods)} methods with metadata.')

metadata = utils.get_metadata(
    names=methods,
    attributes=ATTRIBUTES,
    module_basepath='composer.algorithms',
)

# add extra keys
for name, md in metadata.items():

    # add github link attribute
    md['_github_link'] = GITHUB_BASE + name

    # test that functional form is importable
    if md['_functional']:
        method_functional = md['_functional']

        if not hasattr(CF, method_functional):
            raise ImportError(f'Unable to import functional form {method_functional}')

        md['_functional'] = '`CF.{}`'.format(md['_functional'])

# define row format
row = [
    '[{_class_name}]({_github_link})',
    '{_functional}',
    lambda d: '[{_attribution}]({_link})' if d['_link'] else d['_attribution'],
    '{_tldr}',
    lambda d: '[Link]({_method_card})' if d['_method_card'] else '',
]

table_md = utils.build_markdown_table(
    header=HEADER,
    metadata=metadata,
    sorted_keys=sorted(metadata.keys()),
    row_format=row,
)

# update table in README.md
source_file = Path(__file__).parent.joinpath('README.md')
utils.update_table_in_file(table_md, source_file)



# Copyright 2021 MosaicML. All Rights Reserved.

import os
import pathlib
from typing import List

import _pytest.config
import _pytest.config.argparsing
import _pytest.fixtures
import _pytest.mark
import pytest

_MAIN_PYTEST_KEY = "MAIN_PYTEST"


@pytest.fixture
def is_main_pytest_process() -> bool:
    if _MAIN_PYTEST_KEY in os.environ:
        return False
    else:
        os.environ[_MAIN_PYTEST_KEY] = "1"
        return True


@pytest.fixture(autouse=True)
def reset_osenviron(is_main_pytest_process: None):
    # Require the is_main_pytest_process fixture so that will be set before the env variable capture
    original_environ = os.environ.copy()
    try:
        yield
    finally:
        # remove the extra keys
        keys_to_remove: List[str] = []
        for k in os.environ.keys():
            if k not in original_environ:
                keys_to_remove.append(k)
        for k in keys_to_remove:
            del os.environ[k]

        # reset the other keys
        for k, v in original_environ.items():
            os.environ[k] = v


@pytest.fixture(autouse=True)
def ddp_fork_multi_pytest(request: _pytest.fixtures.FixtureRequest, reset_osenviron: None) -> None:
    # When using DDP fork, the same command is executed again.
    # When multiple tests are executed with a single pytest command,
    # then the DDP fork will re-run all tests, when instead we just want
    # the forked test to run again. To prevent this behavior, we store the
    # running pytest id as an environ. If this variable is set, and it does
    # NOT match the current pytest test ID, then the test should be skipped
    # as the test was launched by a different DDP fork
    test_name: str = request.node.name
    if "PYTEST_TEST_ID" not in os.environ:
        os.environ["PYTEST_TEST_ID"] = test_name
        return
    if os.environ["PYTEST_TEST_ID"] != test_name:
        pytest.skip("Skipping test invoked within a DDP fork")


@pytest.fixture
def ddp_tmpdir(tmpdir: pathlib.Path) -> str:
    if os.environ.get("DDP_TMP_DIR") is None:
        # Need to set the tmpdir as an environ, so all subprocesses share the same tmpdir.
        # Upon subprocess calls, pytest gives each ddp process a different tmpdir
        # Instead, we want to use the same tmpdir for test case assertions.
        os.environ['DDP_TMP_DIR'] = str(tmpdir)
        return str(tmpdir)
    return os.environ["DDP_TMP_DIR"]



# Copyright 2021 MosaicML. All Rights Reserved.

from copy import deepcopy

import pytest
import torch

from composer.algorithms import LayerFreezing, LayerFreezingHparams
from composer.core.state import State
from composer.core.types import Event, Model, Precision
from composer.loggers import Logger
from composer.trainer.trainer_hparams import TrainerHparams
from composer.utils import ensure_tuple
from tests.utils.trainer_fit import train_model


def _generate_state(epoch: int, max_epochs: int, model: Model):
    state = State(
        epoch=epoch,
        step=epoch,
        train_batch_size=64,
        eval_batch_size=64,
        grad_accum=1,
        max_epochs=max_epochs,
        model=model,
        optimizers=(torch.optim.SGD(model.parameters(), lr=0.01, momentum=0.99),),
        precision=Precision.FP32,
    )
    return state


def _check_param_groups(expected_groups, actual_groups):
    assert len(expected_groups) == len(actual_groups), 'Incorrect number of param groups'

    for i, expected_group in enumerate(expected_groups):
        assert len(expected_group) == len(actual_groups[i]), \
            f'Group {i} has the wrong number of parameters'

        for j, expected_params in enumerate(expected_group['params']):
            torch.testing.assert_equal(actual_groups[i]['params'][j], expected_params)


def test_freeze_layers_no_freeze(simple_conv_model: Model, noop_dummy_logger: Logger):
    state = _generate_state(epoch=10, max_epochs=100, model=simple_conv_model)

    first_optimizer = ensure_tuple(state.optimizers)[0]
    assert first_optimizer is not None

    expected_param_groups = deepcopy(first_optimizer.param_groups)
    freezing = LayerFreezing(freeze_start=0.5, freeze_level=1.0)
    freezing.apply(event=Event.EPOCH_END, state=state, logger=noop_dummy_logger)
    updated_param_groups = first_optimizer.param_groups

    _check_param_groups(expected_param_groups, updated_param_groups)


def test_freeze_layers_with_freeze(simple_conv_model: Model, noop_dummy_logger: Logger):
    state = _generate_state(epoch=80, max_epochs=100, model=simple_conv_model)

    first_optimizer = ensure_tuple(state.optimizers)[0]
    assert first_optimizer is not None

    expected_param_groups = deepcopy(first_optimizer.param_groups)
    # The first group should be removed due to freezing
    expected_param_groups[0]['params'] = expected_param_groups[0]['params'][1:]
    freezing = LayerFreezing(freeze_start=0.05, freeze_level=1.0)
    freezing.apply(event=Event.EPOCH_END, state=state, logger=noop_dummy_logger)
    updated_param_groups = first_optimizer.param_groups

    _check_param_groups(expected_param_groups, updated_param_groups)


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_layer_freezing_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [LayerFreezingHparams(freeze_start=.25, freeze_level=1)]
    train_model(mosaic_trainer_hparams, max_epochs=4)



# Copyright 2021 MosaicML. All Rights Reserved.

import torch


def _attn(self, query, key, value, attention_mask=None, head_mask=None):
    attn_weights = torch.matmul(query, key.transpose(-1, -2))

    if self.scale_attn_weights:
        attn_weights = attn_weights / (float(value.size(-1))**0.5)

    # This is the modification from the original attention
    n_tokens = attn_weights.shape[-1]
    # Truncate alibi distance weights to size of current batch
    alibi = self.alibi[:, :, 0:n_tokens]
    # alibi = self.alibi[:, :, :, 0:n_tokens].repeat(batch_size, 1, 1, 1)
    attn_weights = attn_weights + alibi
    # End modification

    if not self.is_cross_attention:
        # if only "normal" attention layer implements causal mask
        query_length, key_length = query.size(-2), key.size(-2)
        causal_mask = self.bias[:, :, key_length - query_length:key_length, :key_length].bool()
        attn_weights = torch.where(causal_mask, attn_weights, self.masked_bias.to(attn_weights.dtype))

    if attention_mask is not None:
        # Apply the attention mask
        attn_weights = attn_weights + attention_mask

    attn_weights = torch.nn.Softmax(dim=-1)(attn_weights)
    attn_weights = self.attn_dropout(attn_weights)

    # Mask heads if we want to
    if head_mask is not None:
        attn_weights = attn_weights * head_mask

    attn_output = torch.matmul(attn_weights, value)

    return attn_output, attn_weights


def enlarge_mask(module: torch.nn.Module, max_sequence_length: int):
    old_mask = module.bias
    new_mask = torch.tril(
        torch.ones((max_sequence_length, max_sequence_length), dtype=torch.uint8,
                   device=old_mask.device)).view(1, 1, max_sequence_length, max_sequence_length)  # type: ignore
    setattr(module, "bias", new_mask)
    return module
